# -*- coding: utf-8 -*-

DOCUMENT_METADATA_ADD_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataAddAction'
DOCUMENT_METADATA_EDIT_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataEditAction'
DOCUMENT_METADATA_REMOVE_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataRemoveAction'

TEST_DATE_INVALID = '___________'
TEST_DEFAULT_VALUE = 'test'

TEST_LOOKUP_TEMPLATE = '1,2,3'
TEST_LOOKUP_VALUE_CORRECT = '1'
TEST_LOOKUP_VALUE_INCORRECT = '0'

TEST_METADATA_INDEX_NODE_TEMPLATE = '{{ document.metadata.first.metadata_type.label }}-{{ document.metadata.first.value }}'

TEST_METADATA_TYPE_DEFAULT_VALUE = 'default value'
TEST_METADATA_TYPE_LABEL = 'test metadata type label'
TEST_METADATA_TYPE_LABEL_EDITED = 'test metadata type label edited'
TEST_METADATA_TYPE_NAME = 'test_metadata_type_name'
TEST_METADATA_TYPE_NAME_EDITED = 'test_metadata_type_name_edited'
TEST_METADATA_VALUE = 'test value'
TEST_METADATA_VALUE_EDITED = 'test value edited'
TEST_METADATA_VALUE_UNICODE = 'español'
TEST_METADATA_VALUE_WITH_AMPERSAND = 'first value & second value'

TEST_PARSER_DATE_VALID = '2001-01-01'
TEST_PARSER_PATH_DATE = 'mayan.apps.metadata.metadata_parsers.DateParser'

TEST_PARSER_REGULAR_EXPRESSION = 'mayan.apps.metadata.metadata_parsers.RegularExpressionParser'
TEST_PARSER_REGULAR_EXPRESSION_PATTERN = 'abc'
TEST_PARSER_REGULAR_EXPRESSION_REPLACEMENT_TEXT = 'replaced_text'

TEST_VALID_DATE = '2001-1-1'
TEST_VALIDATOR_PATH_DATE = 'mayan.apps.metadata.metadata_validators.DateValidator'
TEST_VALIDATOR_PATH_REGULAR_EXPRESSION = 'mayan.apps.metadata.metadata_validators.RegularExpressionValidator'
TEST_VALIDATOR_REGULAR_EXPRESSION_PATTERN = '^[A-Za-z]*$'
TEST_VALIDATOR_VALUE_INVALID = '1234'
TEST_VALIDATOR_VALUE_VALID = 'abcd'



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0019_workflowtransition_condition')
    ]

    operations = [
        migrations.AddField(
            model_name='workflowstateaction',
            name='condition',
            field=models.TextField(
                blank=True, help_text="The condition that will "
                "determine if this state action is executed or not. "
                "The condition is evaluated against the workflow "
                "instance. Conditions that return None or an empty "
                "string ('') are considered to be logical false, "
                "any other value is considered to be the logical "
                "true.", verbose_name='Condition'
            )
        ),
        migrations.AlterField(
            model_name='workflowtransition',
            name='condition',
            field=models.TextField(
                blank=True, help_text="The condition that will "
                "determine if this transition is enabled or not. "
                "The condition is evaluated against the workflow "
                "instance. Conditions that return None or an empty "
                "string ('') are considered to be logical false, "
                "any other value is considered to be the logical "
                "true.", verbose_name='Condition'
            )
        )
    ]



import logging

from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import DynamicFormModelBackend
from mayan.apps.common.classes import PropertyHelper
from mayan.apps.templating.classes import Template

from .exceptions import WorkflowStateActionError

logger = logging.getLogger(name=__name__)


class DocumentStateHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentStateHelper(*args, **kwargs)

    def get_result(self, name):
        return self.instance.workflows.get(workflow__internal_name=name)


class WorkflowAction(DynamicFormModelBackend):
    _backend_app_label = 'document_states'
    _backend_model_name = 'WorkflowStateAction'
    _loader_module_name = 'workflow_actions'

    form_fieldsets = (
        (
            _(message='General'), {
                'fields': ('label', 'enabled', 'when', 'condition')
            }
        ),
    )

    @classmethod
    def clean(cls, request, form_data=None):
        return form_data

    @classmethod
    def get_choices(cls):
        apps_name_map = {
            app.name: app for app in apps.get_app_configs()
        }

        # Match each workflow action to an app.
        apps_workflow_action_map = {}

        for klass in WorkflowAction.get_all():
            for app_name, app in apps_name_map.items():
                if klass.__module__.startswith(app_name):
                    apps_workflow_action_map.setdefault(
                        app, []
                    )
                    apps_workflow_action_map[app].append(
                        (
                            klass.backend_id, klass.label
                        )
                    )

        result = [
            (app.verbose_name, workflow_actions) for app, workflow_actions in apps_workflow_action_map.items()
        ]

        # Sort by app, then by workflow action.
        return sorted(
            result, key=lambda x: (
                x[0], x[1]
            )
        )

    @classmethod
    def migrate(cls):
        WorkflowStateAction = apps.get_model(
            app_label='document_states', model_name='WorkflowStateAction'
        )
        for previous_dotted_path in cls.previous_dotted_paths:
            try:
                WorkflowStateAction.objects.filter(
                    backend_path=previous_dotted_path
                ).update(
                    backend_path=cls.id()
                )
            except (OperationalError, ProgrammingError):
                # Ignore errors during the database migration and
                # quit further attempts.
                return

    def execute(self, context):
        raise NotImplementedError

    @classmethod
    def get_form_schema(cls, workflow_template_state, **kwargs):
        cls.workflow_template_state = workflow_template_state
        return super().get_form_schema(**kwargs)

    def render_field(self, field_name, context):
        try:
            template_string = self.kwargs.get(field_name, '')
            template = Template(template_string=template_string)
            result = template.render(context=context)
        except Exception as exception:
            raise WorkflowStateActionError(
                _(message='%(field_name)s template error: %(exception)s') % {
                    'field_name': field_name, 'exception': exception
                }
            )

        logger.debug('%s template result: %s', field_name, result)

        return result


class WorkflowActionNull(WorkflowAction):
    label = _(message='Null backend')



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_file,
    search_model_document_file_page
)

# Document

search_model_document.add_model_field(
    field='files__file_pages__content__content',
    label=_(message='Document file content')
)

# Document file

search_model_document_file.add_model_field(
    field='file_pages__content__content', label=_(message='Content')
)

# Document file page

search_model_document_file_page.add_model_field(
    field='content__content', label=_(message='Document file page content')
)



from mayan.apps.documents.tests.base import GenericTransactionDocumentViewTestCase
from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..events import event_document_version_exported
from ..permissions import permission_document_version_export

from .mixins import DocumentVersionExportViewTestMixin


class DocumentVersionExportViewTestCase(
    DocumentVersionExportViewTestMixin,
    GenericTransactionDocumentViewTestCase
):
    """
    Use a transaction test case to test the transaction.on_commit code
    of the export task. Use convert back to a normal test case and use
    `captureOnCommitCallbacks` when upgraded to Django 3.2:
    https://github.com/django/django/commit/e906ff6fca291fc0bfa0d52f05817ee9dae0335d
    """

    def test_document_version_export_view_no_permission(self):
        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_export_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_export
        )

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        response = self._request_test_document_version_export_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count + 1
        )

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_download_file)
        self.assertEqual(events[0].verb, event_download_file_created.id)

        self.assertEqual(events[1].action_object, test_download_file)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document_version)
        self.assertEqual(events[1].verb, event_document_version_exported.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, test_message)
        self.assertEqual(events[2].target, test_message)
        self.assertEqual(events[2].verb, event_message_created.id)

    def test_trashed_document_version_export_view_with_access(self):
        self.grant_access(
            obj=self._test_document_version,
            permission=permission_document_version_export
        )

        download_file_count = DownloadFile.objects.count()

        self._test_document.delete()
        self._clear_events()

        response = self._request_test_document_version_export_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0013_auto_20170206_0710')
    ]

    operations = [
        migrations.AlterField(
            model_name='sanescanner',
            name='adf_mode',
            field=models.CharField(
                blank=True, choices=[
                    ('simplex', 'Simplex'), ('duplex', 'Duplex')
                ], default='simplex', help_text='Selects the document feeder '
                'mode (simplex/duplex). If this option is not supported by '
                'your scanner, leave it blank.', max_length=16,
                verbose_name='ADF mode'
            )
        ),
        migrations.AlterField(
            model_name='sanescanner',
            name='source',
            field=models.CharField(
                blank=True, choices=[
                    ('flatbed', 'Flatbed'),
                    ('"Automatic Document Feeder"', 'Document feeder')
                ], default='flatbed', help_text='Selects the scan source '
                '(such as a document-feeder). If this option is not '
                'supported by your scanner, leave it blank.', max_length=32,
                verbose_name='Paper source'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers

from .models import Key


class KeySerializer(serializers.ModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'key_id',
                'view_name': 'rest_api:key-detail'
            }
        }
        fields = (
            'algorithm', 'creation_date', 'expiration_date', 'fingerprint',
            'id', 'key_data', 'key_type', 'length', 'url', 'user_id'
        )
        model = Key
        read_only_fields = (
            'algorithm', 'creation_date', 'expiration_date', 'fingerprint',
            'id', 'key_type', 'length', 'url', 'user_id'
        )



from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.storage.classes
import mayan.apps.storage.models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('permissions', '0004_auto_20191213_0044'),
        ('storage', '0003_auto_20200601_0649')
    ]

    operations = [
        migrations.CreateModel(
            name='DownloadFile',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'file', models.FileField(
                        storage=mayan.apps.storage.classes.DefinedStorageLazy(
                            name='storage__downloadfile'
                        ),
                        upload_to=mayan.apps.storage.models.download_file_upload_to,
                        verbose_name='File'
                    )
                ),
                (
                    'filename', models.CharField(
                        db_index=True, max_length=255,
                        verbose_name='Filename'
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, verbose_name='Date time'
                    )
                ),
                (
                    'object_id', models.PositiveIntegerField(
                        blank=True, null=True
                    )
                ),
                (
                    'label', models.CharField(
                        db_index=True, max_length=192, verbose_name='Label'
                    )
                ),
                (
                    'content_type', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.ContentType'
                    )
                ),
                (
                    'permission', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='permissions.StoredPermission',
                        verbose_name='Permission'
                    )
                )
            ],
            options={
                'verbose_name': 'Download file',
                'verbose_name_plural': 'Download files'
            }
        )
    ]



import json

from django.core.files.base import ContentFile

from mayan.apps.credentials.tests.mixins import StoredCredentialPasswordUsernameTestMixin
from mayan.apps.events.classes import EventModelRegistry
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..classes import ModelMailingActionAttachment, ModelMailingActionLink
from ..models import UserMailer

from .literals import (
    TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS,
    TEST_MAILING_OBJECT_CONTENT, TEST_MAILING_OBJECT_MIME_TYPE,
    TEST_MAILING_PROFILE_BACKEND_PATH, TEST_MAILING_PROFILE_LABEL,
    TEST_MAILING_PROFILE_LABEL_EDITED
)


class MailingProfileTestMixin(
    StoredCredentialPasswordUsernameTestMixin, TestMixinObjectCreationTrack
):
    _test_mailing_profile_auto_create = False
    _test_mailing_profile_backend_path = TEST_MAILING_PROFILE_BACKEND_PATH
    _test_object_model = UserMailer
    _test_object_name = '_test_mailing_profile'

    def setUp(self):
        super().setUp()

        if self._test_mailing_profile_auto_create:
            self._create_test_mailing_profile()

    def _create_test_mailing_profile(self, extra_backend_data=None):
        backend_data = {'from': TEST_EMAIL_FROM_ADDRESS}

        if extra_backend_data:
            backend_data.update(**extra_backend_data)

        self._test_mailing_profile = UserMailer.objects.create(
            default=True, enabled=True, label=TEST_MAILING_PROFILE_LABEL,
            backend_path=self._test_mailing_profile_backend_path,
            backend_data=json.dumps(obj=backend_data)
        )


class DocumentMailingProfileViewTestMixin(MailingProfileTestMixin):
    def _request_test_document_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_link_single', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            },
        )

    def _request_test_document_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_link_multiple', query={
                'id_list': self._test_document.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )


class DocumentFileMailingProfileViewTestMixin(MailingProfileTestMixin):
    def _request_test_document_file_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_single', kwargs={
                'document_file_id': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_file_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_link_multiple', query={
                'id_list': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_file_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_single', kwargs={
                'document_file_id': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_file_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_file_attachment_multiple', query={
                'id_list': self._test_document_file.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )


class DocumentVersionMailingProfileViewTestMixin(MailingProfileTestMixin):
    def _request_test_document_version_link_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_single', kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_version_link_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_link_multiple', query={
                'id_list': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_version_attachment_send_single_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_single',
            kwargs={
                'document_version_id': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )

    def _request_test_document_version_attachment_send_multiple_view(self):
        return self.post(
            viewname='mailer:send_document_version_attachment_multiple',
            query={
                'id_list': self._test_document_version.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                ),
                'mailing_profile': self._test_mailing_profile.pk
            }
        )


def _test_content_function(obj):
    return ContentFile(content=TEST_MAILING_OBJECT_CONTENT)


def _test_mime_type_function(obj):
    return TEST_MAILING_OBJECT_MIME_TYPE


class MailObjectSendAPIViewTestMixin(MailingProfileTestMixin):
    _test_model_mailing_content_function_dotted_path = 'mayan.apps.mailer.tests.mixins._test_content_function'
    _test_model_mailing_mime_type_function_dotted_path = 'mayan.apps.mailer.tests.mixins._test_mime_type_function'
    _test_model_mailing_permission_attachment = None
    _test_model_mailing_permission_link = None

    def setUp(self):
        super().setUp()

        if self.auto_create_test_object:
            EventModelRegistry.register(model=self.TestModel)

            self.TestModel.add_to_class(
                name='get_absolute_url', value=lambda self: None
            )

            if self._test_model_mailing_permission_attachment:
                ModelMailingActionAttachment(
                    content_function_dotted_path=self._test_model_mailing_content_function_dotted_path,
                    manager_name='objects',
                    mime_type_function_dotted_path=self._test_model_mailing_mime_type_function_dotted_path,
                    model=self.TestModel,
                    permission=self._test_model_mailing_permission_attachment
                )

            if self._test_model_mailing_permission_link:
                ModelMailingActionLink(
                    manager_name='objects',
                    model=self.TestModel,
                    permission=self._test_model_mailing_permission_link
                )

    def _request_object_mail_attachment_api_view(self):
        return self.post(
            viewname='rest_api:object-mailing-action-attachment',
            kwargs=self._test_object_view_kwargs,
            data={
                'mailing_profile': self._test_mailing_profile.pk,
                'recipient': TEST_EMAIL_ADDRESS
            }
        )

    def _request_object_mail_link_api_view(self):
        return self.post(
            viewname='rest_api:object-mailing-action-link',
            kwargs=self._test_object_view_kwargs,
            data={
                'mailing_profile': self._test_mailing_profile.pk,
                'recipient': TEST_EMAIL_ADDRESS
            }
        )


class MailingProfileAPIViewTestMixin(MailingProfileTestMixin):
    def _request_test_mailing_profile_create_api_view(self):
        self._test_object_track()

        response = self.post(
            viewname='rest_api:mailing_profile-list', data={
                'backend_path': TEST_MAILING_PROFILE_BACKEND_PATH,
                'label': TEST_MAILING_PROFILE_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_mailing_profile_delete_api_view(self):
        return self.delete(
            viewname='rest_api:mailing_profile-detail',
            kwargs={'mailing_profile_id': self._test_mailing_profile.pk}
        )

    def _request_test_mailing_profile_detail_api_view(self):
        return self.get(
            viewname='rest_api:mailing_profile-detail',
            kwargs={'mailing_profile_id': self._test_mailing_profile.pk}
        )

    def _request_test_mailing_profile_edit_api_view(
        self, extra_data=None, verb='patch'
    ):
        data = {
            'backend_path': TEST_MAILING_PROFILE_BACKEND_PATH,
            'label': TEST_MAILING_PROFILE_LABEL_EDITED
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, verb)(
            viewname='rest_api:mailing_profile-detail',
            kwargs={'mailing_profile_id': self._test_mailing_profile.pk},
            data=data
        )

    def _request_test_mailing_profile_list_api_view(self):
        return self.get(viewname='rest_api:mailing_profile-list')


class MailingProfileViewTestMixin(MailingProfileTestMixin):
    def _request_test_mailing_profile_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='mailer:mailing_profile_create', kwargs={
                'backend_path': TEST_MAILING_PROFILE_BACKEND_PATH
            }, data={
                'default': True,
                'enabled': True,
                'label': TEST_MAILING_PROFILE_LABEL
            }
        )

        self._test_object_set()

        return response

    def _request_test_mailing_profile_delete_view(self):
        return self.post(
            viewname='mailer:mailing_profile_delete', kwargs={
                'mailing_profile_id': self._test_mailing_profile.pk
            }
        )

    def _request_test_mailing_profile_edit_view(self):
        return self.post(
            viewname='mailer:mailing_profile_edit', kwargs={
                'mailing_profile_id': self._test_mailing_profile.pk
            }, data={
                'label': '{}_edited'.format(TEST_MAILING_PROFILE_LABEL)
            }
        )

    def _request_test_mailing_profile_list_view(self):
        return self.get(
            viewname='mailer:mailing_profile_list'
        )

    def _request_test_mailing_profile_log_entry_view(self):
        return self.get(
            viewname='mailer:mailing_profile_log', kwargs={
                'mailing_profile_id': self._test_mailing_profile.pk
            }
        )

    def _request_test_mailing_profile_test_view(self):
        return self.post(
            viewname='mailer:mailing_profile_test', kwargs={
                'mailing_profile_id': self._test_mailing_profile.pk
            }, data={
                'email': getattr(
                    self, '_test_email_address', TEST_EMAIL_ADDRESS
                )
            }
        )



import logging

from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin

__all__ = ('DuplicateBackend',)
logger = logging.getLogger(name=__name__)


class DuplicateBackendMetaclass(type):
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )
        if not new_class.__module__ == 'mayan.apps.duplicates.classes':
            mcs._registry[
                '{}.{}'.format(new_class.__module__, name)
            ] = new_class

        return new_class


class DuplicateBackend(
    AppsModuleLoaderMixin, metaclass=DuplicateBackendMetaclass
):
    _loader_module_name = 'duplicate_backends'

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_all(cls):
        return cls._registry.items()

    @classmethod
    def get_choices(cls):
        return sorted(
            (
                (
                    key, backend.label
                ) for key, backend in cls.get_all().items()
            ), key=lambda x: x[1]
        )

    @classmethod
    def get_class_path(cls):
        for path, klass in cls.get_all().items():
            if klass is cls:
                return path

    @classmethod
    def verify(cls, document):
        """
        Method to check is the document has can be scanned for duplicates.
        Returns either a true (True or anything) or false value (False, None).
        """
        return True

    def __init__(self, model_instance_id, **kwargs):
        self.model_instance_id = model_instance_id
        self.kwargs = kwargs

    def get_model_instance(self):
        StoredDuplicateBackend = apps.get_model(
            app_label='duplicated', model_name='StoredDuplicateBackend'
        )
        return StoredDuplicateBackend.objects.get(pk=self.model_instance_id)

    def process(self, document):
        logger.info(
            'Starting processing document: %s', document
        )

        return self._process(document=document)

    def _process(self, document):
        raise NotImplementedError(
            'Your %s class has not defined the required '
            '_process() method.' % self.__class__.__name__
        )


class NullBackend(DuplicateBackend):
    label = _(message='Null backend')



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from ..events import (
    event_document_auto_checked_in, event_document_checked_out
)
from ..exceptions import (
    DocumentAlreadyCheckedOut, DocumentNotCheckedOut,
    NewDocumentFileNotAllowed
)
from ..models import DocumentCheckout

from .mixins import DocumentCheckoutTestMixin


class DocumentCheckoutTestCase(
    DocumentCheckoutTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_check_out(self):
        self._clear_events()

        self._check_out_test_document()

        self.assertTrue(
            DocumentCheckout.objects.is_checked_out(
                document=self._test_document
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_checked_out.id)

    def test_document_check_in(self):
        self._check_out_test_document()

        self._clear_events()

        self._test_document.check_in()

        self.assertFalse(
            self._test_document.is_checked_out()
        )
        self.assertFalse(
            DocumentCheckout.objects.is_checked_out(
                document=self._test_document
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_auto_checked_in.id)

    def test_document_double_check_out(self):
        self._create_test_case_super_user()
        self._check_out_test_document()

        self._clear_events()

        with self.assertRaises(expected_exception=DocumentAlreadyCheckedOut):
            DocumentCheckout.objects.check_out_document(
                document=self._test_document,
                expiration_datetime=self._check_out_expiration_datetime,
                user=self._test_case_super_user,
                block_new_file=True
            )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_check_in_without_check_out(self):
        self._clear_events()

        with self.assertRaises(expected_exception=DocumentNotCheckedOut):
            self._test_document.check_in()

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_auto_check_in(self):
        self._check_out_test_document()

        # Ensure we wait from longer than the document check out expiration.
        self._test_delay(seconds=self._test_document_check_out_seconds + 0.1)

        self._clear_events()

        DocumentCheckout.objects.check_in_expired_check_outs()

        self.assertFalse(
            self._test_document.is_checked_out()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_auto_checked_in.id)

    def test_method_get_absolute_url(self):
        self._clear_events()

        self._check_out_test_document()

        self._clear_events()

        self.assertTrue(
            self._test_check_out.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_blocking_new_files(self):
        # Silence unrelated logging.
        self._silence_logger(
            name='mayan.apps.documents.models.document_model_mixins'
        )
        self._check_out_test_document()

        self._clear_events()

        with self.assertRaises(expected_exception=NewDocumentFileNotAllowed):
            with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
                self._test_document.files_upload(file_object=file_object)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_file_creation_blocking(self):
        # Silence unrelated logging.
        self._silence_logger(
            name='mayan.apps.documents.models.document_model_mixins'
        )

        self._create_test_case_super_user()

        self._check_out_test_document()

        self._clear_events()

        with self.assertRaises(expected_exception=NewDocumentFileNotAllowed):
            with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
                self._test_document.files_upload(file_object=file_object)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.document_file_actions import (
    DocumentFileActionAppendNewPages, DocumentFileActionNothing,
    DocumentFileActionUseNewPages
)
from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import (
    TEST_FILE_COMPRESSED_PATH, TEST_DOCUMENT_SMALL_CHECKSUM
)
from mayan.apps.source_compressed.source_backends.literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK,
    SOURCE_UNCOMPRESS_CHOICE_NEVER
)
from mayan.apps.sources.events import event_source_created
from mayan.apps.sources.models import Source
from mayan.apps.sources.permissions import permission_sources_create
from mayan.apps.sources.tests.mixins.source_view_mixins import (
    SourceActionViewTestMixin, SourceViewTestMixin
)

from .mixins import StagingFolderSourceTestMixin


class StagingFolderSourceBackendActionDocumentFileUploadViewTestCase(
    SourceActionViewTestMixin, StagingFolderSourceTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = True

    def test_basic_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_document_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_source_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_full_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )
        self.assertEqual(
            self._test_document.versions.all()[1].pages.count(), 1
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)

    def test_basic_with_full_access_trashed_document(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        self._test_document.delete()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view(
            extra_data={'document-comment': 'test-comment'}
        )
        self.assertEqual(response.status_code, 302)

        self._test_document.refresh_from_db()

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.file_latest.comment, 'test-comment'
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )
        self.assertEqual(
            self._test_document.versions.all()[1].pages.count(), 1
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)

    def test_document_file_action_append(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view(
            extra_data={
                'document-action_name': DocumentFileActionAppendNewPages.action_id
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )
        self.assertEqual(
            self._test_document.versions.all()[1].pages.count(),
            self._test_document.files.all()[0].pages.count() + self._test_document.files.all()[1].pages.count()
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page_first = test_document_version.pages.all()[0]
        test_document_version_page_last = test_document_version.pages.all()[1]

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page_first)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page_last)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_document_file_action_keep(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view(
            extra_data={
                'document-action_name': DocumentFileActionNothing.action_id
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

    def test_document_file_action_new(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_file_new
        )

        test_document_count = Document.objects.count()
        test_document_file_count = self._test_document.files.count()
        test_document_version_count = self._test_document.versions.count()
        test_document_version_page_count = self._test_document.versions.all()[0].pages.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_file_upload_view(
            extra_data={
                'document-action_name': DocumentFileActionUseNewPages.action_id
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), test_document_count
        )
        self.assertEqual(
            self._test_document.files.count(), test_document_file_count + 1
        )
        self.assertEqual(
            self._test_document.versions.count(),
            test_document_version_count + 1
        )
        self.assertEqual(
            self._test_document.versions.all()[0].pages.count(),
            test_document_version_page_count
        )
        self.assertEqual(
            self._test_document.versions.all()[1].pages.count(), 1
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[4].action_object, test_document)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version)
        self.assertEqual(events[4].verb, event_document_version_edited.id)


class StagingFolderSourceBackendActionDocumentUploadViewTestCase(
    SourceActionViewTestMixin, StagingFolderSourceTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_basic_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_document_type_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_source_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_basic_with_full_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_DOCUMENT_SMALL_CHECKSUM
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_basic_compressed_always(self):
        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 2
        )

        self.assertTrue(
            'first document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )
        self.assertTrue(
            'second document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 13)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_documents[0])
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_documents[0].version_active)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, self._test_document_type)
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_documents[1])
        self.assertEqual(events[6].verb, event_document_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_created.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, test_documents[1].file_latest)
        self.assertEqual(events[8].verb, event_document_file_edited.id)

        self.assertEqual(events[9].action_object, test_documents[1])
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(events[9].target, test_documents[1].version_active)
        self.assertEqual(events[9].verb, event_document_version_created.id)

        self.assertEqual(
            events[10].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(
            events[10].target, test_documents[1].version_active.pages.first()
        )
        self.assertEqual(
            events[10].verb, event_document_version_page_created.id
        )

        self.assertEqual(
            events[11].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[11].actor, self._test_case_user)
        self.assertEqual(
            events[11].target, test_documents[1].version_active.pages.last()
        )
        self.assertEqual(
            events[11].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[12].action_object, test_documents[1])
        self.assertEqual(events[12].actor, self._test_case_user)
        self.assertEqual(events[12].target, test_documents[1].version_active)
        self.assertEqual(events[12].verb, event_document_version_edited.id)

    def test_basic_compressed_ask_false(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view(
            extra_data={'source-expand': False}
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)

    def test_basic_compressed_ask_true(self):
        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view(
            extra_data={'source-expand': True}
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 2
        )

        self.assertTrue(
            'first document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )
        self.assertTrue(
            'second document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 13)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_documents[0])
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_documents[0].version_active)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, self._test_document_type)
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_documents[1])
        self.assertEqual(events[6].verb, event_document_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_created.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, test_documents[1].file_latest)
        self.assertEqual(events[8].verb, event_document_file_edited.id)

        self.assertEqual(events[9].action_object, test_documents[1])
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(events[9].target, test_documents[1].version_active)
        self.assertEqual(events[9].verb, event_document_version_created.id)

        self.assertEqual(
            events[10].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(
            events[10].target, test_documents[1].version_active.pages.first()
        )
        self.assertEqual(
            events[10].verb, event_document_version_page_created.id
        )

        self.assertEqual(
            events[11].action_object, test_documents[1].version_active
        )
        self.assertEqual(events[11].actor, self._test_case_user)
        self.assertEqual(
            events[11].target, test_documents[1].version_active.pages.last()
        )
        self.assertEqual(
            events[11].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[12].action_object, test_documents[1])
        self.assertEqual(events[12].actor, self._test_case_user)
        self.assertEqual(events[12].target, test_documents[1].version_active)
        self.assertEqual(events[12].verb, event_document_version_edited.id)

    def test_basic_compressed_never(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        self.copy_test_source_file(source_path=TEST_FILE_COMPRESSED_PATH)

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)

    def test_language(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_document_upload_post_view(
            extra_data={'document-language': 'deu'}
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )
        self.assertEqual(Document.objects.first().language, 'deu')

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)


class StagingFolderSourceBackendActionFileDeleteViewTestCase(
    SourceActionViewTestMixin, StagingFolderSourceTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def _get_test_source_stored_file_list(self):
        return list(
            self._test_source.action_execute(
                interface_name='Model', name='file_list'
            )
        )

    def test_get_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_get_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_get_with_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_get_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._get_test_source_stored_file_list()[0]['encoded_filename'],
            self._test_source_stored_file.encoded_filename
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_post_no_permission(self):
        self._test_source_create()

        self.copy_test_source_file()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_post_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_post_with_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_post_view(
            action_name='file_delete'
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_nonexistent_get_with_access(self):
        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._test_source_stored_test_file.unlink()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_get_view(
            action_name='file_delete'
        )
        self.assertContains(
            response=response, text='Requested file not found',
            status_code=200
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_nonexistent_post_with_access(self):
        self._silence_logger(name='mayan.apps.sources.views.source_views')

        self._test_source_create()

        self.copy_test_source_file()

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._test_source_stored_test_file.unlink()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        response = self._request_test_source_action_post_view(
            action_name='file_delete', follow=True
        )
        self.assertContains(
            response=response, text='Requested file not found',
            status_code=200
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class StagingFolderSourceBackendViewTestCase(
    StagingFolderSourceTestMixin, SourceViewTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_staging_folder_source_create_get_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_get_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_source_create_get_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_source_create_post_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_post_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_source_create_post_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_created.id)



from mayan.apps.document_indexing.models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .literals import (
    TEST_FILE_METADATA_INDEX_NODE_TEMPLATE, TEST_FILE_METADATA_VALUE
)
from .mixins import FileMetadataTestMixin


class IndexingTestCase(
    FileMetadataTestMixin, IndexTemplateTestMixin, GenericDocumentTestCase
):
    _test_document_file_metadata_create_auto = True
    _test_index_template_node_expression = TEST_FILE_METADATA_INDEX_NODE_TEMPLATE
    auto_upload_test_document = False

    def test_indexing(self):
        self._upload_test_document()

        self._test_document.submit_for_file_metadata_processing()

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=TEST_FILE_METADATA_VALUE
            ).exists()
        )



from django.views.generic.list import MultipleObjectMixin

from mayan.apps.views.settings import setting_paginate_by


class SourceBackendStoredFileSourceFileListView(MultipleObjectMixin):
    def get_paginate_by(self, queryset):
        return setting_paginate_by.value



from mayan.apps.document_states.events import event_workflow_template_edited
from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase

from ..events import event_tag_attached, event_tag_removed
from ..models import Tag
from ..permissions import permission_tag_attach
from ..workflow_actions import AttachTagAction, RemoveTagAction

from .mixins import TagTestMixin


class TagWorkflowActionTestCase(
    WorkflowTemplateStateActionTestMixin, TagTestMixin, BaseTestCase
):
    auto_create_test_tag = True

    def test_tag_attach_action(self):
        self._clear_events()

        self._execute_workflow_template_state_action(
            klass=AttachTagAction, kwargs={
                'tags': Tag.objects.all()
            }
        )

        self.assertEqual(
            self._test_tag.documents.count(), 1
        )
        self.assertTrue(
            self._test_document in self._test_tag.documents.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_attached.id)

    def test_tag_remove_action(self):
        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self._clear_events()

        self._execute_workflow_template_state_action(
            klass=RemoveTagAction, kwargs={
                'tags': Tag.objects.all()
            }
        )

        self.assertEqual(
            self._test_tag.documents.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_removed.id)


class TagWorkflowActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, TagTestMixin,
    GenericViewTestCase
):
    auto_create_test_tag = True

    def test_tag_attach_action_create_get_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        self.grant_access(
            obj=self._test_tag,
            permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.tags.workflow_actions.AttachTagAction'
        )
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_attach_action_create_post_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.tags.workflow_actions.AttachTagAction',
            extra_data={
                'tags': self._test_tag.pk
            }
        )
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_workflow_template_state_action
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_tag_remove_action_create_get_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path='mayan.apps.tags.workflow_actions.RemoveTagAction'
        )
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_remove_action_create_post_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.tags.workflow_actions.RemoveTagAction',
            extra_data={
                'tags': self._test_tag.pk
            }
        )
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_workflow_template_state_action
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from .classes import DocumentFileAction


class DocumentFileActionAppendNewPages(DocumentFileAction):
    action_id = 'append'
    label = _(message='Append. Create a new version and append the new file pages.')

    @staticmethod
    def execute(document, document_file, comment, user):
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        annotated_content_object_list = []
        annotated_content_object_list.extend(
            DocumentVersion.annotate_content_object_list(
                content_object_list=document.version_active.page_content_objects
            )
        )

        annotated_content_object_list.extend(
            DocumentVersion.annotate_content_object_list(
                content_object_list=document_file.pages.all(),
                start_page_number=document.version_active.pages.count() + 1
            )
        )

        document_version = DocumentVersion(
            active=True, document=document, comment=comment
        )
        document_version._event_actor = user
        document_version.save()

        document_version.pages_remap(
            annotated_content_object_list=annotated_content_object_list,
            user=user
        )


class DocumentFileActionNothing(DocumentFileAction):
    action_id = 'keep'
    label = _(
        'Keep. Do not create a new version and keep the current '
        'version pages.'
    )

    @staticmethod
    def execute(document, document_file, comment, user):
        return


class DocumentFileActionUseNewPages(DocumentFileAction):
    action_id = 'replace'
    label = _(message='Replace. Create a new version and use the new file pages.')

    @staticmethod
    def execute(document, document_file, comment, user):
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        document_version = DocumentVersion(
            active=True, document=document, comment=comment
        )
        document_version._event_actor = user
        document_version.save()

        annotated_content_object_list = DocumentVersion.annotate_content_object_list(
            content_object_list=document_file.pages.all()
        )

        document_version.pages_remap(
            annotated_content_object_list=annotated_content_object_list,
            user=user
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Authentication'), name='authentication'
)

permission_users_impersonate = namespace.add_permission(
    label=_(message='Impersonate users'), name='users_impersonate'
)



import uuid

import django.core.files.storage
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0041_auto_20170823_1855')
    ]

    operations = [
        migrations.AlterField(
            model_name='documentversion',
            name='file',
            field=models.FileField(
                storage=django.core.files.storage.FileSystemStorage(
                    location=b'mayan/media/document_storage'
                ), upload_to=str(
                    uuid.uuid4()
                ), verbose_name='File'
            )
        )
    ]



import json

from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)
from mayan.apps.user_management.serializers import UserSerializer

from ..models import WorkflowInstance, WorkflowInstanceLogEntry
from ..permissions import permission_workflow_tools

from .workflow_template_serializers import WorkflowTemplateSerializer
from .workflow_template_state_serializers import WorkflowTemplateStateSerializer
from .workflow_template_transition_serializers import WorkflowTemplateTransitionSerializer


class WorkflowInstanceLaunchSerializer(serializers.Serializer):
    workflow_template_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the workflow template to launch.'
        ), label=_(message='Workflow template ID'),
        source_permission=permission_workflow_tools
    )

    def get_workflow_template_id_queryset(self):
        return self.context['document_type'].workflows.exclude(
            id__in=self.context['document'].workflows.values('id')
        )


class WorkflowInstanceLogEntrySerializer(serializers.ModelSerializer):
    document_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Document URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_instance.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
        ), view_name='rest_api:document-detail'
    )
    transition = WorkflowTemplateTransitionSerializer(
        label=_(message='Transition'), read_only=True
    )
    transition_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the transition to be added.'
        ), label=_(message='Transition ID'),
        source_queryset_method='get_workflow_instance_transition_queryset',
        write_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_instance.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'workflow_instance.pk',
                'lookup_url_kwarg': 'workflow_instance_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_instance_log_entry_id'
            }
        ), view_name='rest_api:workflow-instance-log-entry-detail'
    )
    user = UserSerializer(
        label=_(message='User'), read_only=True
    )
    workflow_instance_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow instance URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_instance.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'workflow_instance.pk',
                'lookup_url_kwarg': 'workflow_instance_id'
            }
        ), view_name='rest_api:workflow-instance-detail'
    )
    workflow_template_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow template URL'), view_kwargs=(
            {
                'lookup_field': 'workflow_instance.workflow.pk',
                'lookup_url_kwarg': 'workflow_template_id'
            },
        ), view_name='rest_api:workflow-template-detail'
    )

    class Meta:
        fields = (
            'comment', 'datetime', 'document_url', 'extra_data', 'id',
            'transition', 'transition_id', 'url', 'user',
            'workflow_instance_url', 'workflow_template_url'
        )
        model = WorkflowInstanceLogEntry
        read_only_fields = (
            'datetime', 'document_url', 'id', 'transition',
            'url', 'user', 'workflow_instance_url', 'workflow_template_url'
        )

    def create(self, validated_data):
        return self.context['workflow_instance'].do_transition(
            transition=validated_data['transition_id'],
            comment=validated_data.get('comment'),
            extra_data=json.loads(
                s=validated_data.get('extra_data', '{}')
            ), user=self.context['request'].user
        )

    def get_workflow_instance_transition_queryset(self):
        return self.context['workflow_instance'].get_transition_choices(
            user=self.context['request'].user
        )


class WorkflowInstanceSerializer(serializers.ModelSerializer):
    workflow_template = WorkflowTemplateSerializer(
        label=_(message='Workflow template'), read_only=True, source='workflow'
    )
    context = serializers.SerializerMethodField(
        label=_(message='Context')
    )
    current_state = WorkflowTemplateStateSerializer(
        label=_(message='Current state'), read_only=True, source='get_current_state'
    )
    document_url = serializers.SerializerMethodField(
        label=_(message='Document URL')
    )
    last_log_entry = WorkflowInstanceLogEntrySerializer(
        label=_(message='Last log entry'), read_only=True,
        source='get_last_log_entry'
    )
    log_entries_url = serializers.SerializerMethodField(
        label=_(message='Log entries URL'), help_text=_(
            'A link to the entire history of this workflow.'
        )
    )
    log_entry_transitions_url = serializers.SerializerMethodField(
        label=_(message='Log entry transitions URL'), read_only=True
    )
    # DEPRECATION: Remove in version 5.0.
    workflow_template_url = serializers.SerializerMethodField(
        label=_(message='Workflow template URL')
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL'), help_text=_(
            'API URL pointing to a workflow in relation to the '
            'document to which it is attached. This URL is different than '
            'the canonical workflow URL.'
        )
    )

    class Meta:
        fields = (
            'context', 'current_state', 'document_url', 'id',
            'last_log_entry', 'log_entries_url',
            'log_entry_transitions_url', 'url', 'workflow_template',
            'workflow_template_url'
        )
        model = WorkflowInstance
        read_only_fields = fields

    def get_document_url(self, instance):
        return reverse(
            viewname='rest_api:document-detail', kwargs={
                'document_id': instance.document.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_context(self, instance):
        return {
            'workflow_instance_context': instance.loads()
        }

    def get_log_entries_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-instance-log-entry-list', kwargs={
                'document_id': instance.document.pk,
                'workflow_instance_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_log_entry_transitions_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-instance-log-entry-transition-list',
            kwargs={
                'document_id': instance.document.pk,
                'workflow_instance_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-instance-detail', kwargs={
                'document_id': instance.document.pk,
                'workflow_instance_id': instance.pk
            }, request=self.context['request'], format=self.context['format']
        )

    def get_workflow_template_url(self, instance):
        return reverse(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': instance.workflow.pk
            }, request=self.context['request'], format=self.context['format']
        )



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='ErrorLog',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'name', models.CharField(
                        max_length=128, verbose_name='Internal name'
                    )
                )
            ],
            options={
                'verbose_name': 'Error log',
                'verbose_name_plural': 'Error logs',
                'ordering': ('name',)
            }
        ),
        migrations.CreateModel(
            name='ErrorLogPartition',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'name', models.CharField(
                        db_index=True, max_length=128,
                        verbose_name='Internal name'
                    )
                ),
                (
                    'error_log', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='partitions', to='logging.ErrorLog',
                        verbose_name='Error log'
                    )
                )
            ],
            options={
                'verbose_name': 'Error log partition',
                'verbose_name_plural': 'Error log partitions',
                'unique_together': {
                    ('error_log', 'name')
                }
            }
        ),
        migrations.CreateModel(
            name='ErrorLogPartitionEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, db_index=True,
                        verbose_name='Date and time'
                    )
                ),
                (
                    'text', models.TextField(
                        blank=True, null=True, verbose_name='Text'
                    )
                ),
                (
                    'error_log_partition', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='entries',
                        to='logging.ErrorLogPartition',
                        verbose_name='Error log partition'
                    )
                )
            ],
            options={
                'verbose_name': 'Error log partition entry',
                'verbose_name_plural': 'Error log partition entries',
                'ordering': ('datetime',),
                'get_latest_by': 'datetime'
            }
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_document_downloads = CeleryQueue(
    name='documents_downloads', label=_(message='Documents downloads'),
    worker=worker_c
)

queue_document_downloads.add_task_type(
    dotted_path='mayan.apps.document_downloads.tasks.task_document_file_compress',
    label=_(message='Generate a compressed document file bundle')
)



import logging

from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.views import (
    ViewSingleObjectDynamicFormModelBackendCreate,
    ViewSingleObjectDynamicFormModelBackendEdit
)
from mayan.apps.views.generics import (
    FormView, SingleObjectDeleteView, SingleObjectListView
)

from .classes import CredentialBackend
from .forms import (
    StoredCredentialBackendDynamicForm, StoredCredentialBackendSelectionForm
)
from .icons import (
    icon_credential_backend_selection, icon_credential_create,
    icon_credential_delete, icon_credential_edit, icon_credential_list,
    icon_credentials
)
from .links import link_credential_backend_selection
from .models import StoredCredential
from .permissions import (
    permission_credential_create, permission_credential_delete,
    permission_credential_edit, permission_credential_view
)

logger = logging.getLogger(name=__name__)


class StoredCredentialBackendSelectionView(FormView):
    extra_context = {
        'title': _(message='New credential backend selection')
    }
    form_class = StoredCredentialBackendSelectionForm
    view_icon = icon_credential_backend_selection
    view_permission = permission_credential_create

    def form_valid(self, form):
        backend = form.cleaned_data['backend']
        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='credentials:stored_credential_create', kwargs={
                    'backend_path': backend
                }
            )
        )


class StoredCredentialCreateView(ViewSingleObjectDynamicFormModelBackendCreate):
    backend_class = CredentialBackend
    form_class = StoredCredentialBackendDynamicForm
    post_action_redirect = reverse_lazy(
        viewname='credentials:stored_credential_list'
    )
    view_icon = icon_credential_create
    view_permission = permission_credential_create

    def get_extra_context(self):
        backend_class = self.get_backend_class()

        return {
            'title': _(
                'Create a "%s" credential'
            ) % backend_class.label
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            '_event_ignore_credential_used': True,
            'backend_path': self.kwargs['backend_path']
        }


class StoredCredentialDeleteView(SingleObjectDeleteView):
    model = StoredCredential
    object_permission = permission_credential_delete
    pk_url_kwarg = 'stored_credential_id'
    post_action_redirect = reverse_lazy(
        viewname='credentials:stored_credential_list'
    )
    view_icon = icon_credential_delete

    def get_extra_context(self):
        return {
            'title': _(message='Delete credential: %s') % self.object
        }


class StoredCredentialEditView(ViewSingleObjectDynamicFormModelBackendEdit):
    form_class = StoredCredentialBackendDynamicForm
    model = StoredCredential
    object_permission = permission_credential_edit
    pk_url_kwarg = 'stored_credential_id'
    view_icon = icon_credential_edit

    def get_extra_context(self):
        return {
            'title': _(message='Edit credential: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            '_event_ignore_credential_used': True
        }


class StoredCredentialListView(SingleObjectListView):
    model = StoredCredential
    object_permission = permission_credential_view
    view_icon = icon_credential_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_credentials,
            'no_results_main_link': link_credential_backend_selection.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Credentials represent an identity. '
                'These are used to when accessing external systems or '
                'services.'
            ),
            'no_results_title': _(message='No credentials available'),
            'title': _(message='Credentials')
        }



from rest_framework import status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from mayan.apps.acls.models import AccessControlList
from mayan.apps.rest_api import generics

from ..api_view_mixins import ParentObjectSourceAPIViewMixin
from ..exceptions import SourceActionException
from ..permissions import permission_sources_view
from ..serializers import SourceBackendActionSerializer


class APISourceActionDetailView(
    ParentObjectSourceAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Source action detail view.
    """
    lookup_url_kwarg = 'action_name'
    mayan_external_object_permission_map = {'GET': permission_sources_view}
    serializer_class = SourceBackendActionSerializer

    def get_object(self):
        for entry in self.get_queryset():
            if entry.name == self.kwargs['action_name']:
                return entry

    def get_source_queryset(self):
        return self.get_source().get_action_list()


class APISourceActionExecuteView(
    ParentObjectSourceAPIViewMixin, generics.ObjectActionAPIView
):
    """
    get: Execute a source action without confirmation.
    post: Execute a source action with confirmation.
    """
    lookup_url_kwarg = 'action_name'
    serializer_class = SourceBackendActionSerializer

    def get(self, request, *args, **kwargs):
        if self.get_object().confirmation:
            handler = self.http_method_not_allowed
            response = handler(request, *args, **kwargs)
            self.response = self.finalize_response(
                request, response, *args, **kwargs
            )
            return self.response
        else:
            return self.view_action(request, *args, **kwargs)

    def get_object(self):
        source = self.get_source()

        action = source.get_action(
            name=self.kwargs['action_name']
        )

        action_permission = action.permission

        # Filter the source again if the action has a permission requirement.

        if action_permission:
            source_queryset = self.get_parent_queryset_source()

            queryset = AccessControlList.objects.restrict_queryset(
                permission=action_permission, queryset=source_queryset,
                user=self.request.user
            )

            get_object_or_404(
                queryset=queryset, pk=self.kwargs['source_id']
            )

        return action

    def get_parent_queryset_source(self):
        return super().get_parent_queryset_source().filter(enabled=True)

    def get_serializer_extra_context(self):
        return {
            'action': self.get_object()
        }

    def get_source_queryset(self):
        # Does nothing. Required by the API view class filtering.
        source = self.get_source()

        return source.get_action_list()

    def post(self, request, *args, **kwargs):
        if self.get_object().confirmation:
            return self.view_action(request, *args, **kwargs)
        else:
            handler = self.http_method_not_allowed
            response = handler(request, *args, **kwargs)
            self.response = self.finalize_response(
                request, response, *args, **kwargs
            )
            return self.response

    def view_action(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        action = self.get_object()

        interface_kwargs = {}

        serializer_arguments = serializer.validated_data.get(
            'arguments', {}
        )

        interface_kwargs.update(**request.GET)
        interface_kwargs.update(**serializer_arguments)

        if action.accept_files:
            interface_kwargs['file'] = serializer.validated_data.get('file')

        interface_kwargs['request'] = request

        try:
            result = action.execute(
                interface_name='RESTAPI',
                interface_load_kwargs=interface_kwargs,
                interface_retrieve_kwargs={'view': self}
            )
        except SourceActionException as exception:
            return Response(
                data=str(exception), exception=exception,
                status=status.HTTP_400_BAD_REQUEST
            )
        else:
            return result


class APISourceActionListView(
    ParentObjectSourceAPIViewMixin, generics.ListAPIView
):
    """
    get: Source action list view.
    """
    mayan_external_object_permission_map = {'GET': permission_sources_view}
    serializer_class = SourceBackendActionSerializer

    def get_source_queryset(self):
        return list(
            self.get_source().get_action_list()
        )



from django.contrib.contenttypes.models import ContentType

from mayan.apps.rest_api import generics

from .serializers import ContentTypeSerializer


class APIContentTypeDetailView(generics.RetrieveAPIView):
    """
    Returns the details of the selected content type.
    """
    lookup_url_kwarg = 'content_type_id'
    model = ContentType
    serializer_class = ContentTypeSerializer


class APIContentTypeListView(generics.ListAPIView):
    """
    Returns a list of all the available content types.
    """
    serializer_class = ContentTypeSerializer
    source_queryset = ContentType.objects.order_by('app_label', 'model')



from django.db import migrations, models, reset_queries
import django.db.models.deletion


def code_remap_document_version_pages_ocr_content(apps, schema_editor):
    """
    The document_page_id field is pointing to the document file page.
    This migration will remap that to point it for the corresponding
    document version page that is connected to the original document file
    page.
    """
    cursor_primary = schema_editor.connection.create_cursor(
        name='merged_content_page_id'
    )
    cursor_secondary = schema_editor.connection.cursor()

    query = '''
        SELECT
            {ocr_documentpageocrcontent}.{content},
            {documents_documentversionpage}.{id}
        FROM {ocr_documentpageocrcontent}
        LEFT OUTER JOIN
            {documents_documentversionpage} ON (
                {documents_documentversionpage}.{object_id} = {ocr_documentpageocrcontent}.{document_page_id}
            )
    '''.format(
        content=schema_editor.connection.ops.quote_name(name='content'),
        document_page_id=schema_editor.connection.ops.quote_name(
            name='document_page_id'
        ),
        documents_documentversionpage=schema_editor.connection.ops.quote_name(
            name='documents_documentversionpage'
        ),
        id=schema_editor.connection.ops.quote_name(name='id'),
        object_id=schema_editor.connection.ops.quote_name(name='object_id'),
        ocr_documentpageocrcontent=schema_editor.connection.ops.quote_name(
            name='ocr_documentpageocrcontent'
        )
    )

    cursor_primary.execute(query)

    insert_query = '''
        INSERT INTO {ocr_documentversionpageocrcontent} (
            content,document_version_page_id
        ) VALUES {{}};
    '''.format(
        ocr_documentversionpageocrcontent=schema_editor.connection.ops.quote_name(
            name='ocr_documentversionpageocrcontent'
        )
    )

    FETCH_SIZE = 10000

    while True:
        rows = cursor_primary.fetchmany(FETCH_SIZE)

        if not rows:
            break

        insert_query_final = insert_query.format(
            (
                '(%s,%s),' * len(rows)
            )[:-1]
        )

        tuples = []
        for row in rows:
            tuples += row

        cursor_secondary.execute(insert_query_final, tuples)
        reset_queries()


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0075_delete_duplicateddocumentold'),
        ('ocr', '0008_auto_20180917_0646')
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentVersionPageOCRContent',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'content', models.TextField(
                        blank=True, help_text='The actual text content '
                        'extracted by the OCR backend.',
                        verbose_name='Content'
                    )
                ),
                (
                    'document_version_page', models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='ocr_content',
                        to='documents.DocumentVersionPage',
                        verbose_name='Document version page'
                    )
                )
            ],
            options={
                'verbose_name': 'Document version page OCR content',
                'verbose_name_plural': 'Document version pages OCR contents'
            }
        ),
        migrations.RunPython(
            code=code_remap_document_version_pages_ocr_content,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.shortcuts import get_object_or_404

from mayan.apps.acls.models import AccessControlList

from ..models.index_instance_models import IndexInstance
from ..permissions import permission_index_instance_view
from ..serializers import IndexInstanceNodeSerializer


class APIIndexInstanceNodeViewMixin:
    serializer_class = IndexInstanceNodeSerializer

    def get_index_instance(self):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_index_instance_view,
            queryset=IndexInstance.objects.all(), user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['index_instance_id']
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Cabinets'), name='cabinets'
)


event_cabinet_created = namespace.add_event_type(
    label=_(message='Cabinet created'), name='cabinet_created'
)
event_cabinet_deleted = namespace.add_event_type(
    label=_(message='Cabinet deleted'), name='cabinet_deleted'
)
event_cabinet_edited = namespace.add_event_type(
    label=_(message='Cabinet edited'), name='cabinet_edited'
)
event_cabinet_document_added = namespace.add_event_type(
    label=_(message='Document added to cabinet'), name='add_document'
)
event_cabinet_document_removed = namespace.add_event_type(
    label=_(message='Document removed from cabinet'), name='remove_document'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.literals import DEFAULT_DOCUMENT_FILE_ACTION_NAME
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_models import Document

from ..interface_arguments import SourceBackendActionInterfaceArgument

from .argument_help_texts import (
    argument_help_text_document, argument_help_text_immediate_mode
)
from .literals import DEFAULT_IMMEDIATE_MODE

argument_document = SourceBackendActionInterfaceArgument(
    help_text=argument_help_text_document
)
argument_document_description = SourceBackendActionInterfaceArgument(
    default=None, help_text=Document.description.field.help_text,
    required=False
)
argument_document_label = SourceBackendActionInterfaceArgument(
    default=None, help_text=Document.label.field.help_text, required=False
)
argument_document_language = SourceBackendActionInterfaceArgument(
    default=None, help_text=Document.language.field.help_text, required=False
)
argument_document_id = SourceBackendActionInterfaceArgument(
    help_text=_(
        'ID of the document to which a new file will be uploaded to.'
    )
)
argument_document_id_optional = SourceBackendActionInterfaceArgument(
    default=None, help_text=argument_help_text_document, required=False
)

# Document file

argument_document_file_action_name = SourceBackendActionInterfaceArgument(
    default=DEFAULT_DOCUMENT_FILE_ACTION_NAME,
    help_text=DocumentFile.versions_new.help_text, required=False
)
argument_document_file_comment = SourceBackendActionInterfaceArgument(
    default=None, help_text=DocumentFile.comment.field.help_text,
    required=False
)
argument_document_file_filename = SourceBackendActionInterfaceArgument(
    default=None, help_text=DocumentFile.filename.field.help_text,
    required=False
)

# Document type

argument_document_type = SourceBackendActionInterfaceArgument(
    help_text=Document.description.field.help_text
)
argument_document_type_id = SourceBackendActionInterfaceArgument(
    help_text=_(message='ID of the document type.')
)

# Immediate mode

argument_immediate_mode_optional = SourceBackendActionInterfaceArgument(
    default=DEFAULT_IMMEDIATE_MODE,
    help_text=argument_help_text_immediate_mode, required=False
)
argument_immediate_mode_required = SourceBackendActionInterfaceArgument(
    help_text=argument_help_text_immediate_mode
)

# User

argument_user = SourceBackendActionInterfaceArgument(
    default=None, help_text=_(
        'User that will feature as the actor in the events.'
    ), required=False
)
argument_user_id = SourceBackendActionInterfaceArgument(
    default=None, help_text=_(
        'ID of the user that will feature as the actor in '
        'the events.'
    ), required=False
)



from mayan.apps.appearance.classes import Icon
from mayan.apps.documents.icons import icon_document, icon_document_type

icon_workflow = Icon(driver_name='fontawesome', symbol='sitemap')

# Tools

icon_tool_launch_workflows = Icon(
    driver_name='fontawesome-dual', primary_symbol='sitemap',
    secondary_symbol='play'
)
icon_document_workflow_templates_launch = icon_tool_launch_workflows

# Workflow template

icon_workflow_template_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='sitemap',
    secondary_symbol='plus'
)
icon_workflow_template_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_document_type_list = icon_document_type
icon_workflow_template_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_launch = Icon(
    driver_name='fontawesome', symbol='play'
)
icon_workflow_template_list = icon_workflow
icon_workflow_template_preview = Icon(
    driver_name='fontawesome', symbol='eye'
)

# Document type

icon_document_type_workflow_template_list = icon_workflow

# Workflow instance

icon_workflow_instance_detail = icon_workflow
icon_workflow_instance_list = icon_workflow
icon_workflow_instance_transition = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)
icon_workflow_instance_transition_select = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)

# Workflow runtime proxy

icon_workflow_runtime_proxy_document_list = icon_document
icon_workflow_runtime_proxy_list = icon_workflow
icon_workflow_runtime_proxy_state_document_list = icon_document
icon_workflow_runtime_proxy_state_list = Icon(
    driver_name='fontawesome', symbol='circle'
)

# Workflow transition state

icon_workflow_template_state_action_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_state_action_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_state_action_list = Icon(
    driver_name='fontawesome', symbol='code'
)
icon_workflow_template_state = Icon(
    driver_name='fontawesome', symbol='circle'
)
icon_workflow_template_state_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='circle',
    secondary_symbol='plus'
)
icon_workflow_template_state_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_state_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_state_list = Icon(
    driver_name='fontawesome', symbol='circle'
)

# Workflow template state action

icon_workflow_template_state_action = Icon(
    driver_name='fontawesome', symbol='code'
)
icon_workflow_template_state_action_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='code',
    secondary_symbol='plus'
)
icon_workflow_template_state_action_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_state_action_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_state_action_selection = Icon(
    driver_name='fontawesome-dual', primary_symbol='code',
    secondary_symbol='plus'
)
icon_workflow_template_state_action_list = Icon(
    driver_name='fontawesome', symbol='code'
)

# Workflow template state transitions

icon_workflow_template_state_escalation = Icon(
    driver_name='fontawesome', symbol='hourglass'
)
icon_workflow_template_state_escalation_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='hourglass',
    secondary_symbol='plus'
)
icon_workflow_template_state_escalation_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_state_escalation_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_state_escalation_list = Icon(
    driver_name='fontawesome', symbol='hourglass'
)

# Workflow transition

icon_workflow_template_transition = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)
icon_workflow_template_transition_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='arrows-alt-h',
    secondary_symbol='plus'
)
icon_workflow_template_transition_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_transition_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_transition_list = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)

# Workflow transition field

icon_workflow_template_transition_field = Icon(
    driver_name='fontawesome', symbol='table'
)
icon_workflow_template_transition_field_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='table',
    secondary_symbol='plus'
)
icon_workflow_template_transition_field_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_template_transition_field_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_template_transition_field_list = Icon(
    driver_name='fontawesome', symbol='table'
)
icon_workflow_template_transition_triggers = Icon(
    driver_name='fontawesome', symbol='bolt'
)



from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.source_interactive.source_backend_actions.callback_mixins import (
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendActionMixinCallbackDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.base import SourceBackendAction
from mayan.apps.sources.source_backend_actions.mixins.document_file_mixins import SourceBackendActionMixinDocumentFileUploadInteractive
from mayan.apps.sources.source_backend_actions.mixins.document_mixins import (
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionMixinDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.mixins.document_type_mixins import SourceBackendActionMixinDocumentTypeInteractive

from .mixins import SourceBackendActionMixinFileGenerated


class SourceBackendActionGenerateFileBase(
    SourceBackendActionMixinFileGenerated, SourceBackendAction
):
    """
    Base action class for all generated file source actions.
    """


class SourceBackendActionGenerateFileDocumentFileUpload(
    SourceBackendActionMixinDocumentFileUploadInteractive,
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionGenerateFileBase
):
    name = 'document_file_upload'
    permission = permission_document_file_new


class SourceBackendActionGenerateFileDocumentUpload(
    SourceBackendActionMixinDocumentUploadInteractive,
    SourceBackendActionMixinCallbackDocumentUploadInteractive,
    SourceBackendActionMixinDocumentTypeInteractive,
    SourceBackendActionGenerateFileBase
):
    name = 'document_upload'
    permission = permission_document_create



from django import forms
from django.forms.formsets import formset_factory
from django.utils.translation import gettext_lazy as _

from ..fields import DocumentVersionPageField, ThumbnailFormField

from .document_version_page_form_mixins import FormSetExtraFormKwargsMixin


class DocumentVersionPageForm(forms.Form):
    document_version_page = DocumentVersionPageField()

    def __init__(self, *args, **kwargs):
        instance = kwargs.pop('instance', None)
        transformation_instance_list = kwargs.pop(
            'transformation_instance_list', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['document_version_page'].initial = instance
        self.fields['document_version_page'].widget.attrs.update(
            {
                'transformation_instance_list': transformation_instance_list
            }
        )


class DocumentVersionPageMappingForm(forms.Form):
    source_content_type = forms.IntegerField(
        label=_(message='Content type'), widget=forms.HiddenInput
    )
    source_object_id = forms.IntegerField(
        label=_(message='Object ID'), widget=forms.HiddenInput
    )
    source_label = forms.CharField(
        label=_(message='Source'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    source_thumbnail = ThumbnailFormField(required=False)
    target_page_number = forms.ChoiceField(
        choices=(), label=_(message='Destination page number'), required=False,
        widget=forms.widgets.Select(
            attrs={'size': 1, 'class': 'select2'}
        )
    )

    def __init__(self, *args, **kwargs):
        target_page_number_choices = kwargs.pop(
            'target_page_number_choices', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['target_page_number'].choices = target_page_number_choices


class DocumentVersionPageMappingFormSet(
    FormSetExtraFormKwargsMixin, formset_factory(
        form=DocumentVersionPageMappingForm, extra=0
    )
):
    """
    Combined formset
    """
    def clean(self):
        set_of_target_page_numbers = set()
        for form in self.forms:
            cleaned_data_entry = form.cleaned_data
            target_page_number = cleaned_data_entry['target_page_number']
            if target_page_number != '0':
                if target_page_number in set_of_target_page_numbers:
                    form.add_error(
                        error=_(message='Target page number can\'t be repeated.'),
                        field='target_page_number'
                    )
                else:
                    set_of_target_page_numbers.add(target_page_number)



from django.utils.translation import gettext_lazy as _

DOCUMENT_VERSION_EXPORT_MESSAGE_BODY = _(
    'Document version "%(document_version)s" has been '
    'exported and is available for download using the '
    'link: %(download_url)s or from '
    'the downloads area (%(download_list_url)s).'
)
DOCUMENT_VERSION_EXPORT_MESSAGE_SUBJECT = _(message='Document version exported.')
DOCUMENT_VERSION_EXPORT_MIMETYPE = 'application/pdf'



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.exceptions import ActionError
from mayan.apps.views.generics import (
    MultipleObjectConfirmActionView, MultipleObjectFormActionView,
    SingleObjectDetailView
)

from .exceptions import DocumentAlreadyCheckedOut, DocumentNotCheckedOut
from .forms import DocumentCheckOutDetailForm, DocumentCheckOutForm
from .icons import (
    icon_check_in_document, icon_check_out_document, icon_check_out_info,
    icon_check_out_list
)
from .models import DocumentCheckout
from .permissions import (
    permission_document_check_in, permission_document_check_in_override,
    permission_document_check_out, permission_document_check_out_detail_view
)


class DocumentCheckInView(MultipleObjectConfirmActionView):
    error_message = _(
        'Unable to check in document "%(instance)s"; %(exception)s'
    )
    pk_url_kwarg = 'document_id'
    success_message_plural = _(
        '%(count)d documents checked in successfully.'
    )
    success_message_single = _(
        'Document "%(object)s" checked in successfully.'
    )
    success_message_singular = _(
        '%(count)d document checked in successfully.'
    )
    title_plural = _(message='Check in %(count)d documents.')
    title_single = _(message='Check in document "%(object)s".')
    title_singular = _(message='Check in %(count)d document.')
    view_icon = icon_check_in_document

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_post_object_action_url(self):
        if self.action_count == 1:
            return reverse(
                viewname='checkouts:check_out_info', kwargs={
                    'document_id': self.action_id_list[0]
                }
            )
        else:
            super().get_post_action_redirect()

    def get_source_queryset(self):
        # object_permission is None to disable restricting queryset mixin
        # and restrict the queryset ourselves from two permissions.
        queryset_documents = Document.valid.all()

        queryset_check_ins = AccessControlList.objects.restrict_queryset(
            permission=permission_document_check_in,
            queryset=queryset_documents, user=self.request.user
        )

        queryset_check_in_overrides = AccessControlList.objects.restrict_queryset(
            permission=permission_document_check_in_override,
            queryset=queryset_documents, user=self.request.user
        )

        return queryset_check_ins | queryset_check_in_overrides

    def object_action(self, form, instance):
        try:
            DocumentCheckout.business_logic.check_in_document(
                document=instance, user=self.request.user
            )
        except DocumentNotCheckedOut as exception:
            raise ActionError(exception)


class DocumentCheckOutView(MultipleObjectFormActionView):
    error_message = _(
        'Unable to checkout document "%(instance)s"; %(exception)s'
    )
    form_class = DocumentCheckOutForm
    object_permission = permission_document_check_out
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message_plural = _(
        '%(count)d documents checked out successfully.'
    )
    success_message_single = _(
        'Document "%(object)s" checked out successfully.'
    )
    success_message_singular = _(
        '%(count)d document checked out successfully.'
    )
    title_plural = _(message='Checkout %(count)d documents.')
    title_single = _(message='Checkout document "%(object)s".')
    title_singular = _(message='Checkout %(count)d document.')
    view_icon = icon_check_out_document

    def get_extra_context(self):
        context = {}

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first()
                }
            )

        return context

    def get_post_object_action_url(self):
        if self.action_count == 1:
            return reverse(
                viewname='checkouts:check_out_info', kwargs={
                    'document_id': self.action_id_list[0]
                }
            )
        else:
            super().get_post_action_redirect()

    def object_action(self, form, instance):
        try:
            DocumentCheckout.objects.check_out_document(
                block_new_file=form.cleaned_data['block_new_file'],
                document=instance,
                expiration_datetime=form.cleaned_data['expiration_datetime'],
                user=self.request.user
            )
        except DocumentAlreadyCheckedOut as exception:
            raise ActionError(exception)


class DocumentCheckOutDetailView(SingleObjectDetailView):
    form_class = DocumentCheckOutDetailForm
    object_permission = permission_document_check_out_detail_view
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    view_icon = icon_check_out_info

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(
                'Check out details for document: %s'
            ) % self.object
        }


class DocumentCheckOutListView(DocumentListView):
    object_permission = permission_document_check_out_detail_view
    view_icon = icon_check_out_list

    def get_document_queryset(self):
        queryset = DocumentCheckout.objects.checked_out_documents()
        return Document.valid.filter(
            pk__in=queryset.values('pk')
        )

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_icon': icon_check_out_info,
                'no_results_text': _(
                    'Checking out a document, blocks certain operations '
                    'for a predetermined amount of time.'
                ),
                'no_results_title': _(message='No documents have been checked out'),
                'title': _(message='Checked out documents')
            }
        )
        return context



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0015_auto_20190429_1922')
    ]

    operations = [
        migrations.AlterField(
            model_name='index',
            name='label',
            field=models.CharField(
                help_text='Short description of this index.',
                max_length=128, unique=True, verbose_name='Label'
            ),
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Quotas'), name='quotas'
)

event_quota_created = namespace.add_event_type(
    label=_(message='Quota created'), name='quota_created'
)
event_quota_edited = namespace.add_event_type(
    label=_(message='Quota edited'), name='quota_edited'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0078_alter_documenttype_label')
    ]

    operations = [
        migrations.AddField(
            model_name='documentfile', name='size',
            field=models.PositiveIntegerField(
                blank=True, db_index=True, editable=False,
                help_text='The size of the file in bytes.', null=True,
                verbose_name='Size'
            )
        )
    ]



from unittest import skip

from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import (
    COMMAND_NAME_SEARCH_INDEX_OBJECTS, COMMAND_NAME_SEARCH_REINDEX,
    COMMAND_NAME_SEARCH_STATUS
)
from ..search_backends import SearchBackend

from .mixins.backend_mixins import BackendSearchTestMixin
from .mixins.base import TestSearchObjectSimpleTestMixin


class SearchReindexManagementCommandTestCaseMixin(
    BackendSearchTestMixin, ManagementCommandTestMixin,
    TestSearchObjectSimpleTestMixin
):
    _test_management_command_name = COMMAND_NAME_SEARCH_REINDEX

    def _create_test_search_objects(self):
        self._create_test_object(
            instance_kwargs={'char': 'abc'}
        )

    def test_artifacts(self):
        self._test_search_backend.reset()

        queryset = self._do_search(
            query={
                'char': self._test_objects[0].char
            }
        )
        self.assertTrue(
            self._test_objects[0] not in queryset
        )

        self._call_test_management_command()

        queryset = self._do_search(
            query={
                'char': self._test_objects[0].char
            }
        )
        self.assertTrue(
            self._test_objects[0] in queryset
        )

    def test_calling(self):
        self._call_test_management_command()


class DjangoSearchReindexManagementCommandTestCase(
    SearchReindexManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""

    @skip(reason='Backend does not support reindexing.')
    def test_artifacts(self):
        """Backend does not support indexing."""


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchReindexManagementCommandTestCase(
    SearchReindexManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchReindexManagementCommandTestCase(
    SearchReindexManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""


class SearchIndexObjectManagementCommandTestCaseMixin(
    BackendSearchTestMixin, ManagementCommandTestMixin,
    TestSearchObjectSimpleTestMixin
):
    _test_management_command_name = COMMAND_NAME_SEARCH_INDEX_OBJECTS

    def _create_test_search_objects(self):
        self._create_test_object(
            instance_kwargs={'char': 'abc'}
        )
        self._create_test_object(
            instance_kwargs={'char': 'xyz'}
        )

    def test_artifacts(self):
        queryset = self._do_search(
            query={'char': self._test_objects[0].char}
        )
        self.assertTrue(
            self._test_objects[0] in queryset
        )

        queryset = self._do_search(
            query={'char': self._test_objects[1].char}
        )
        self.assertTrue(
            self._test_objects[1] in queryset
        )

        backend = SearchBackend.get_instance()
        backend.reset()

        queryset = self._do_search(
            query={
                'char': self._test_objects[0].char
            }
        )
        self.assertTrue(
            self._test_objects[0] not in queryset
        )

        queryset = self._do_search(
            query={'char': self._test_objects[1].char}
        )
        self.assertTrue(
            self._test_objects[1] not in queryset
        )

        self._call_test_management_command(
            self._test_search_model.full_name, self._test_objects[0].pk
        )

        queryset = self._do_search(
            query={
                'char': self._test_objects[0].char
            }
        )
        self.assertTrue(
            self._test_objects[0] in queryset
        )

        queryset = self._do_search(
            query={
                'char': self._test_objects[1].char
            }
        )
        self.assertTrue(
            self._test_objects[1] not in queryset
        )

        self._call_test_management_command(
            self._test_search_model.full_name, self._test_objects[1].pk
        )

        queryset = self._do_search(
            query={
                'char': self._test_objects[0].char
            }
        )
        self.assertTrue(
            self._test_objects[0] in queryset
        )

        queryset = self._do_search(
            query={
                'char': self._test_objects[1].char
            }
        )
        self.assertTrue(
            self._test_objects[1] in queryset
        )

    def test_calling(self):
        self._call_test_management_command(
            self._test_search_model.full_name, self._test_objects[0].pk
        )


class DjangoSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""

    @skip(reason='Backend does not support indexing.')
    def test_artifacts(self):
        """Backend does not support indexing."""


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""


class SearchStatusManagementCommandTestCaseMixin(
    ManagementCommandTestMixin, TestSearchObjectSimpleTestMixin
):
    _test_management_command_name = COMMAND_NAME_SEARCH_STATUS

    def _create_test_search_objects(self):
        self._create_test_object(
            instance_kwargs={'char': 'abc'}
        )
        self._create_test_object(
            instance_kwargs={'char': 'xyz'}
        )

    def test_artifacts(self):
        stdout, stderr = self._call_test_management_command()

        count = 0

        for line in stdout.split('\n'):
            if self._test_search_model.model_name in line.lower():
                model_name, count = line.split(':')
                count = int(count)

        self.assertEqual(
            count, len(self._test_objects)
        )


class DjangoSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against DjangoSearch backend."""


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against WhooshSearch backend."""



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('ocr', '0010_auto_20210304_1215')
    ]

    operations = [
        migrations.DeleteModel(
            name='DocumentVersionOCRError',
        )
    ]



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..permissions import permission_settings_edit, permission_settings_view

from .literals import (
    TEST_SETTING_VALIDATION_BAD_VALUE, TEST_SETTING_VALIDATION_GOOD_VALUE
)
from .mixins import (
    SettingClusterViewTestMixin, SettingNamespaceViewTestMixin,
    SettingViewTestMixin
)
from .mocks import test_validation_function


class SettingClusterViewTestCase(
    SettingClusterViewTestMixin, GenericViewTestCase
):
    def test_cluster_configuration_save_view_no_permission(self):
        response = self._request_cluster_configuration_save_view()
        self.assertEqual(response.status_code, 403)

    def test_cluster_configuration_save_view_with_permission(self):
        self.grant_permission(permission=permission_settings_edit)

        response = self._request_cluster_configuration_save_view()
        self.assertEqual(response.status_code, 302)


class SettingNamespaceViewTestCase(
    SettingNamespaceViewTestMixin, GenericViewTestCase
):
    _test_setting_namespace_create_auto = True

    def test_namespace_detail_view_no_permission(self):
        response = self._request_namespace_detail_view()
        self.assertEqual(response.status_code, 403)

    def test_namespace_detail_view_with_permission(self):
        self.grant_permission(permission=permission_settings_view)

        response = self._request_namespace_detail_view()
        self.assertEqual(response.status_code, 200)

    def test_namespace_list_view_no_permission(self):
        response = self._request_namespace_list_view()
        self.assertEqual(response.status_code, 403)

    def test_namespace_list_view_with_permission(self):
        self.grant_permission(permission=permission_settings_view)

        response = self._request_namespace_list_view()
        self.assertEqual(response.status_code, 200)


class SettingViewTestCase(SettingViewTestMixin, GenericViewTestCase):
    _test_setting_namespace_create_auto = True

    def test_setting_edit_view_no_permission(self):
        self._create_test_setting()

        test_setting_value = self._test_setting.get_value_current()

        response = self._request_setting_edit_view(
            value=TEST_SETTING_VALIDATION_GOOD_VALUE
        )
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            self._test_setting.get_value_current(), test_setting_value
        )

    def test_setting_edit_view_with_permission(self):
        self.grant_permission(permission=permission_settings_edit)

        self._create_test_setting()

        test_setting_value = self._test_setting.get_value_current()

        response = self._request_setting_edit_view(
            value=TEST_SETTING_VALIDATION_GOOD_VALUE
        )
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            self._test_setting.get_value_current(), test_setting_value
        )

    def test_setting_revert_view_no_permission(self):
        self._create_test_setting()

        test_setting_value = self._test_setting.get_value_current()

        self._test_setting.do_value_set(value='edited')

        response = self._request_setting_revert_view()
        self.assertEqual(response.status_code, 403)

        self.assertNotEqual(
            self._test_setting.get_value_current(), test_setting_value
        )

    def test_setting_revert_view_with_permission(self):
        self.grant_permission(permission=permission_settings_edit)

        self._create_test_setting()

        test_setting_value = self._test_setting.get_value_current()

        self._test_setting.do_value_set(value='edited')

        response = self._request_setting_revert_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_setting.get_value_current(), test_setting_value
        )

    def test_setting_validation_with_permission(self):
        self.grant_permission(permission=permission_settings_edit)

        self._create_test_setting(
            validation_function=test_validation_function
        )

        test_setting_value = self._test_setting.value

        response = self._request_setting_edit_view(
            value=TEST_SETTING_VALIDATION_BAD_VALUE
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self._test_setting.value, test_setting_value
        )



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import (
    TEST_DOCUMENT_SMALL_CHECKSUM, TEST_FILE_SMALL_PATH
)

from .literals import TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC
from .mixins.base_mixins import SourceTestMixin


class SourceBackendDocumentUploadSourceMetadataTestCase(
    SourceTestMixin, GenericDocumentTestCase
):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_DOCUMENT_BASIC
    auto_upload_test_document = False

    def test_basic(self):
        self._clear_events()

        with open(file=TEST_FILE_SMALL_PATH, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={
                    'document_type': self._test_document_type,
                    'file_object': file_object
                }
            )

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(
            test_document_file.checksum, TEST_DOCUMENT_SMALL_CHECKSUM
        )

        self.assertEqual(test_document_file.source_metadata.count(), 1)

        self.assertEqual(
            test_document_file.source_metadata.get(key='source_id').value,
            str(self._test_source.pk)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)



from itertools import chain, islice
import logging
import re

from django.apps import apps
from django.utils.translation import gettext_lazy as _

from .literals import (
    DOCUMENT_FILE_SOURCE_METADATA_BATCH_SIZE,
    REGULAR_EXPRESSION_MATCH_EVERYTHING, REGULAR_EXPRESSION_MATCH_NOTHING
)

logger = logging.getLogger(name=__name__)


class SourceBackendMixinRegularExpression:
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'include_regex': {
                    'class': 'django.forms.CharField',
                    'default': '',
                    'help_text': _(
                        'Regular expression used to select which files '
                        'to upload.'
                    ),
                    'label': _(message='Include regular expression'),
                    'required': False
                },
                'exclude_regex': {
                    'class': 'django.forms.CharField',
                    'default': '',
                    'help_text': _(
                        'Regular expression used to exclude which files '
                        'to upload.'
                    ),
                    'label': _(message='Exclude regular expression'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Content selection'), {
                    'fields': ('include_regex', 'exclude_regex')
                }
            ),
        )

        return fieldsets

    def get_regex_exclude(self):
        return re.compile(
            pattern=self.kwargs.get(
                'exclude_regex', REGULAR_EXPRESSION_MATCH_NOTHING
            ) or REGULAR_EXPRESSION_MATCH_NOTHING
        )

    def get_regex_include(self):
        return re.compile(
            pattern=self.kwargs.get(
                'include_regex', REGULAR_EXPRESSION_MATCH_EVERYTHING
            )
        )


class SourceBackendMixinSourceMetadata:
    def callback_post_document_file_create(
        self, source_metadata=None, **kwargs
    ):
        super().callback_post_document_file_create(**kwargs)

        DocumentFileSourceMetadata = apps.get_model(
            app_label='sources', model_name='DocumentFileSourceMetadata'
        )

        document_file = kwargs['document_file']
        source_id = kwargs['source_id']
        source_metadata = source_metadata or {}

        document_file_source_metadata_list = (
            DocumentFileSourceMetadata(
                document_file=document_file, key=key, source_id=source_id,
                value=value
            ) for key, value in source_metadata.items()
        )

        document_file_source_metadata_source_id = DocumentFileSourceMetadata(
            document_file=document_file, key='source_id',
            source_id=source_id, value=source_id
        )

        document_file_source_metadata_list = chain(
            (document_file_source_metadata_source_id,),
            document_file_source_metadata_list
        )

        while True:
            batch = list(
                islice(
                    document_file_source_metadata_list,
                    DOCUMENT_FILE_SOURCE_METADATA_BATCH_SIZE
                )
            )

            if not batch:
                break

            DocumentFileSourceMetadata.objects.bulk_create(
                batch_size=DOCUMENT_FILE_SOURCE_METADATA_BATCH_SIZE,
                objs=batch
            )



from mayan.apps.converter.transformations import TransformationResize
from mayan.apps.navigation.column_widgets import SourceColumnWidget

from .settings import setting_thumbnail_height, setting_thumbnail_width


class ThumbnailWidget(SourceColumnWidget):
    template_name = 'documents/widgets/thumbnail.html'

    def disable_condition(self, instance):
        return instance.is_in_trash

    def get_extra_context(self):
        transformation_instance_list = (
            TransformationResize(
                width=setting_thumbnail_width.value,
                height=setting_thumbnail_height.value
            ),
        )

        return {
            # Disable the clickable link if the document is in the trash
            'disable_title_link': self.disable_condition(instance=self.value),
            'gallery_name': 'document_list',
            'instance': self.value,
            'transformation_instance_list': transformation_instance_list
        }
